# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import logging
import time
from abc import ABCMeta, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import wait
from itertools import islice
from random import shuffle

import mxnet as mx
import numpy as np

__author__ = 'Jinho D. Choi'


class NLPModel(metaclass=ABCMeta):
    def __init__(self, state, batch_size):
        # label
        """

        :param state:
        :type state: Callable[[NLPGraph, NLPLexiconMapper, bool], NLPState]
        :param batch_size:
        :type batch_size: int
        """
        self.index_map = {}
        # Dict[str, int]
        self.labels = []
        # List[str]

        # init
        self.mxmod = None
        self.state = state
        self.batch_size = batch_size
        self.pool_feature_vector = None

    # ============================== Label ==============================

    def get_label_index(self, label):
        """

        :param label:
        :type label: str
        :return: the index of the label.
        :rtype: int
        """
        return self.index_map.get(label, -1)

    def get_label(self, index):
        """

        :param index: the index of the label to be returned.
        :type index: Union[int, np.int]
        :return: the index'th label.
        :rtype: str
        """
        return self.labels[index]

    def add_label(self, label):
        """

        :param label:
        :type label: str
        :return: the index of the label. Add a label to this map if not exist already.
        :rtype int
        """
        idx = self.get_label_index(label)
        if idx < 0:
            idx = len(self.labels)
            self.index_map[label] = idx
            self.labels.append(label)
        return idx

    @property
    def num_label(self):

        """

        :return:
        :rtype: int
        """
        return len(self.labels)

    # ============================== Feature ==============================

    @abstractmethod
    def x(self, state):
        """

        :param state:
        :type state: NLPState
        :return: the feature vector for the current state.
        :rtype: np.array
        """

    def feature_vectors(self, states):
        """

        :param states:
        :type states: List[NLPState]
        :return:
        :rtype: np.array
        """
        xs_0 = [[self.x(state)[0]] for state in states]
        xs_1 = [[self.x(state)[1]] for state in states]
        out_0 = np.stack(xs_0, axis=0)
        out_1 = np.stack(xs_1, axis=0)
        out_0 = np.squeeze(out_0, axis=1)
        out_1 = np.squeeze(out_1, axis=1)
        return [out_0, out_1]

    def train_instances(self, states, num_threads=1):
        """

        :param states:
        :type states: List[NLPState]
        :param num_threads:
        :type num_threads: int
        :return:
        :rtype: Tuple[np.array, np.array]
        """

        def instances(thread_id=0, batch_size=len(states)):
            bidx = thread_id * batch_size
            eidx = min((thread_id + 1) * batch_size, len(states))
            return zip(*[(self.x(state), state.gold) for state in islice(states, bidx, eidx)])

        def xys(future):
            """

            :param future:
            :type future: Future
            :return:
            """
            xs, ys = future.result()
            return xs, np.array([self.add_label(y) for y in ys])

        if num_threads == 1:
            xs, ys = instances()
            xs_1, xs_2 = zip(*xs)
            x1 = np.stack(xs_1, axis=0)
            x2 = np.stack(xs_2, axis=0)
            data = [x1, x2]
            label = np.array([self.add_label(y) for y in ys])
            return data, label
        pool = ThreadPoolExecutor(num_threads)
        size = np.math.ceil(len(states) / num_threads)
        futures = [pool.submit(instances, i, size) for i in range(num_threads)]
        while wait(futures)[1]:
            pass
        xxs, yys = zip(*[xys(future) for future in futures])
        return np.vstack(xxs), np.hstack(yys)

    # ============================== Module ==============================

    def bind(self, data, label=None, batch_size=32, for_training=True, force_rebind=True):
        """

        :param data:
        :type data: np.array
        :param label:
        :type label: np.array
        :param batch_size:
        :type batch_size: int
        :param for_training:
        :type for_training: bool
        :param force_rebind:
        :type force_rebind: bool
        :return:
        :rtype: mx.io.DataIter
        """
        batches = self.data_iter(data, label, batch_size)
        # mx.io.NDArrayIter
        label_shapes = None if label is None else batches.provide_label
        self.mxmod.bind(data_shapes=batches.provide_data,
                        label_shapes=label_shapes,
                        for_training=for_training,
                        force_rebind=force_rebind)
        return batches

    def fit(self, batches, num_epoch=1):
        """

        :param batches:
        :type batches: mx.io.DataIter
        :param num_epoch:
        :type num_epoch: int
        :return:
        :rtype: np.array
        """
        for _ in range(num_epoch):
            for batch in batches:
                self.mxmod.forward(batch)

                # extract pooled layer feature vector.
                # [0] index output is softmax layer output
                temp_symbol_1 = self.mxmod.get_outputs()[1]
                fea = temp_symbol_1.asnumpy()

                # create combined feature vector
                if self.pool_feature_vector is None:
                    self.pool_feature_vector = fea
                else:
                    self.pool_feature_vector = np.concatenate((self.pool_feature_vector, fea),
                                                              axis=0)

                self.mxmod.backward()
                self.mxmod.update()

            # sync aux params across devices
            arg_params, aux_params = self.mxmod.get_params()
            self.mxmod.set_params(arg_params, aux_params)

            # end of 1 epoch, reset the data-iter for another epoch
            batches.reset()

        return self.predict(batches)

    def predict(self, batches):
        """

        :param batches:
        :type batches: mx.io.DataIter
        :return:
        :rtype: np.array
        """
        return self.mxmod.predict(batches)[0].asnumpy()

    def train(self, trn_graphs, dev_graphs, lexicon, num_steps=2000, bagging_ratio=0.63,
              initializer=mx.initializer.Normal(0.01), arg_params=None, aux_params=None,
              allow_missing=False, force_init=False, kvstore='local', optimizer='Adam',
              optimizer_params=(('learning_rate', 0.01),)):

        """

        :param trn_graphs:
        :type trn_graphs: List[NLPGraph]
        :param dev_graphs:
        :type dev_graphs: List[NLPGraph]
        :param lexicon:
        :type lexicon: NLPLexiconMapper
        :param num_steps:
        :type num_steps: int
        :param bagging_ratio:
        :type bagging_ratio: int
        :param initializer:
        :type initializer:  mx.initializer.Initializer
        :param arg_params:
        :type arg_params:
        :param aux_params:
        :type aux_params:
        :param allow_missing:
        :type allow_missing: bool
        :param force_init:
        :type force_init: bool
        :param kvstore:
        :type kvstore: Union[str, mx.kvstore.KVStore]
        :param optimizer:
        :type optimizer: Union[str, mx.optimizer.Optimizer]
        :param optimizer_params:
        :type optimizer_params:
        """
        trn_states = [self.state(graph, lexicon, save_gold=True) for graph in trn_graphs]
        # List[NLPState]
        dev_states = [self.state(graph, lexicon, save_gold=True) for graph in dev_graphs]
        # List[NLPState]

        bag_size = int(len(trn_states) * bagging_ratio)

        best_eval = 0
        for step in range(1, num_steps + 1):
            st = time.time()
            shuffle(trn_states)
            trn_states.sort(key=lambda x: x.reset_count)
            xs, ys = self.train_instances(trn_states[:bag_size])
            batches = self.bind(xs, ys, batch_size=self.batch_size)

            if step == 1:
                self.mxmod.init_params(
                    initializer=initializer, arg_params=arg_params, aux_params=aux_params,
                    allow_missing=allow_missing, force_init=force_init)
                self.mxmod.init_optimizer(
                    kvstore=kvstore, optimizer=optimizer, optimizer_params=optimizer_params)

            predictions = self.fit(batches)
            correct = 0

            for state, y, yhats in zip(trn_states, ys, predictions):
                yh = np.argmax(yhats if len(yhats) == self.num_label else yhats[:self.num_label])
                state.process(self.get_label(yh), yhats)
                if y == yh:
                    correct += 1
                if state.terminate:
                    state.reset()

            trn_acc = correct / len(ys)
            dev_eval = self.evaluate(dev_states, batch_size=self.batch_size)
            tt = time.time() - st
            logging.info('%6d: trn-acc = %6.4f, dev-eval = %6.4f, time = %d',
                         step, trn_acc, dev_eval, tt)
            best_eval = max(dev_eval, best_eval)

        logging.info('best: %6.4f', best_eval)

    def evaluate(self, states, batch_size=32):
        """

        :param states:
        :type states:  List[NLPState]
        :param batch_size:
        :type batch_size: int
        :return:
        """
        for state in states:
            state.reset()
        backup = states

        while states:
            xs = self.feature_vectors(states)
            batches = self.bind(xs, batch_size=batch_size, for_training=False)
            # mx.io.NDArrayIter
            predictions = self.predict(batches)

            for state, yhats in zip(states, predictions):
                yh = np.argmax(yhats if len(yhats) == self.num_label else yhats[:self.num_label])
                state.process(self.get_label(yh), yhats)

            states = [state for state in states if not state.terminate]

        stats = np.array([0, 0])
        acc = 0

        for state in backup:
            acc = state.eval(stats)

        return acc

    # ============================== Helper ==============================

    @classmethod
    def data_iter(cls, data, label=None, batch_size=32):
        """

        :param data:
        :type data: np.array
        :param label:
        :type label: np.array
        :param batch_size:
        :type batch_size: int
        :return:
        :rtype: mx.io.DataIter
        """
        batch_size = len(data[0]) if len(data[0]) < batch_size else batch_size
        return mx.io.NDArrayIter(data={'data_f2v': data[0], 'data_a2v': data[1]}, label=label,
                                 batch_size=batch_size, shuffle=False)



from elit.tokenizer import english_tokenizer
import os
import unittest
#
# RESOURCE_PATH = './../elit/resources/tokenizer'
#
#
# def resource_file(filename):
#     return "{0}/{1}".format(RESOURCE_PATH, filename)
#
#
# class EmoticonsTest(unittest.TestCase):
#
#     def setUp(self):
#         # self.file = os.path.join(os.path.abspath(os.path.dirname(__file__)),
#         #                          resource_file('EMOTICONS'))
#         pass
#
#     def test_tokenizer(self):
#         test_case = ':smile: :hug: :pencil:'
#         tokens = english_tokenizer.tokenize(test_case)
#         self.assertEqual(len(tokens), 3)
#
#
# if __name__ == '__main__':
#     unittest.main()



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import os
import ujson
from elit.lexicon import Word2Vec
from elit.sentiment.sentiment_analyzer import SSTSentimentAnalyzer, TwitterSentimentAnalyzer

__author__ = 'Jinho D. Choi'

import json
from enum import Enum
from elit.tokenizer import english_tokenizer

DOC_MAX_SIZE = 10485760
DOC_DELIM = '@#DOC$%'

FLAG_INPUT_FORMAT = 0
FLAG_TOKENIZATION = 1
FLAG_SEGMENTATION = 2
FLAG_SENTIMENT = 3

KEY_TOKENS = 'tokens'
KEY_OFFSETS = 'offsets'
KEY_SENTIMENT = 'sentiment'
KEY_SENTIMENT_ATTENTION = 'sentiment_attention'

class Language(Enum):
    English = 'en'


class ELITDecoder:
    def __init__(self, resource_dir, language=Language.English, sentiment_twitter=True, sentiment_movie=True):
        if language == Language.English:
            # tokenizer
            english_tokenizer.init(os.path.join(resource_dir, 'tokenizer'))
            self.tokenize = english_tokenizer.tokenize
            self.segment = english_tokenizer.segment

            # init sentiment: twitter
            if sentiment_twitter:
                self.emb_twitter = Word2Vec(os.path.join(resource_dir, 'embedding/w2v-400-twitter.gnsm'))
                self.sentiment_twitter = TwitterSentimentAnalyzer(
                    self.emb_twitter, os.path.join(resource_dir, 'sentiment/sentiment-semeval17-400-v2'))

            # init sentiment: movie review
            if sentiment_movie:
                self.emb_review = Word2Vec(os.path.join(resource_dir, 'embedding/w2v-400-amazon-review.gnsm'))
                self.sentiment_sst = SSTSentimentAnalyzer(
                    self.emb_review, os.path.join(resource_dir, 'sentiment/sentiment-sst-400-v2'))
        else:
            raise ValueError('Unsupported language: '+str(language))

    ############################## DECODE ##############################

    def decode(self, flag, istream, ostream=None):
        """
        :param flag:
        :param istream: either StringIO or File
        :param ostream: either StringIO or File
        :return:
        """
        if ostream is not None:
            ostream.write('[')

        d = self.decode_raw(flag, istream, ostream) if flag[FLAG_INPUT_FORMAT] == '0' \
            else self.decode_line(flag, istream, ostream)

        if ostream is not None:
            ostream.write(']')
        return d

    def decode_raw(self, flag, istream, ostream=None):
        def decode():
            d = self.text_to_sentences(flag, ''.join(lines))
            if ostream is None:
                documents.append(d)
            else:
                ostream.write(str(json.dumps(d)) + ',')

        documents = []
        offset = 0
        lines = []

        for line in istream:
            if line.strip() == DOC_DELIM:
                decode()
                offset = 0
                lines.clear()
            elif offset + len(line) <= DOC_MAX_SIZE:
                offset += len(line)
                lines.append(line)

        if lines:
            decode()
        return documents

    def decode_line(self, flag, istream, ostream=None):
        def decode():
            if ostream is None:
                documents.append(sentences)
            else:
                ostream.write(str(json.dumps(sentences)) + ',')

        documents = []
        sentences = []
        offset = 0

        for line in istream:
            if line.strip() == DOC_DELIM:
                decode()
                offset = 0
                sentences = []
            elif offset + len(line) <= DOC_MAX_SIZE:
                d = self.text_to_sentences(flag, line, offset)
                offset += len(line)
                sentences.extend(d)

        if sentences:
            decode()
        return documents

    ############################## CONVERSION ##############################

    def text_to_sentences(self, flag, text, offset=0):
        tokens = self.tokenize(text, flag[FLAG_TOKENIZATION] == '0')

        sentences = [self.tokens_to_sentence(flag, tokens, offset)] if flag[FLAG_SEGMENTATION] == '0' else \
                     self.tokens_to_sentences(flag, tokens, self.segment(tokens), offset)

        if flag[FLAG_SENTIMENT] != '0':
            self.sentiment_analyze(int(flag[FLAG_SENTIMENT]), sentences)

        return sentences

    def tokens_to_sentence(self, flag, tokens, offset):
        sentence = {KEY_TOKENS: [token[0] for token in tokens],
                    KEY_OFFSETS: [(token[1] + offset, token[2] + offset) for token in tokens]}

        return sentence

    def tokens_to_sentences(self, flag, tokens, segments, offset):
        return [self.tokens_to_sentence(flag, tokens[segments[i]:segments[i + 1]], offset) for i in
                range(0, len(segments) - 1)]

    ############################## COMPONENTS ##############################

    def sentiment_analyze(self, flag, sentences):
        analyzer = self.sentiment_twitter if flag % 2 == 1 else self.sentiment_sst
        attn = flag > 2

        sens = [sentence[KEY_TOKENS] for sentence in sentences]
        y, att, raw_att = analyzer.decode(sens, attn=attn)

        for i, sentence in enumerate(sentences):
            sentence[KEY_SENTIMENT] = y[i].tolist()
            if attn: sentence[KEY_SENTIMENT_ATTENTION] = att[i].tolist()


# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
from fasttext.model import WordVectorModel
from gensim.models import KeyedVectors

from elit.dev.structure import NLPGraph
from elit.dev.template.lexicon import NLPLexiconMapper
from elit.dev.template.state import NLPState

__author__ = 'Jinho D. Choi'

# transitions
LEFT_ARC  = 'L'
RIGHT_ARC = 'R'
NO_ARC    = 'N'
SHIFT     = 'S'
REDUCE    = 'R'
PASS      = 'P'


class DEPLexicon(NLPLexiconMapper):
    def __init__(self, w2v: KeyedVectors=None, f2v: WordVectorModel=None):
        super().__init__(w2v, f2v)


class DEPState(NLPState):
    def __init__(self, graph: NLPGraph, lexicon: DEPLexicon, save_gold=False):
        super().__init__(graph)
        self.lex: DEPLexicon = lexicon

        # reset
        self.stack: List[int] = [0]
        self.inter: List[int] = []
        self.input: int = 1

    # ============================== Node ==============================

    def get_stack(self, window: int=0, relation: Relation=None) -> NLPNode:
        """
        :param window: the context window to the top of the stack.
        :param relation: the relation to the (top+window)'th node.
        :return: relation(top+window)'th node if exists; otherwise, None.
        """
        return self.get_node(index=self.stack[-1], window=window, relation=relation, root=True)

    def get_input(self, window: int=0, relation: Relation=None) -> NLPNode:
        """
        :param window: the context window to the front of the input.
        :param relation: the relation to the (input+window)'th node.
        :return: relation(input+window)'th node if exists; otherwise, None.
        """
        return self.get_node(index=self.input, window=window, relation=relation, root=True)

    # ============================== Transition ==============================

    def next(self, label: str):
        """
        :param label: arc + transition + label
        """
        s = self.get_stack()
        i = self.get_input()
        a = label[0]  # arc
        t = label[1]  # transition

        if a == LEFT_ARC:
            s.set_parent(i, label[2:])
            if t == REDUCE: self.reduce()
            else: self.passes()
        elif a == RIGHT_ARC:
            i.set_parent(s, label[2:])
            if t == SHIFT: self.shift()
            else: self.passes()
        else:
            if t == SHIFT: self.shift()
            elif t == REDUCE: self.reduce()
            else: self.passes()

    def shift(self):
        self.stack.extend(reversed(self.inter))
        self.stack.append(self.input)
        del self.inter[:]
        self.input += 1

    def reduce(self):
        self.stack.pop()

    def passes(self):
        self.inter.append(self.stack.pop())

    def terminate(self):
        return self.input >= len(self.graph)

    # ============================== Node ==============================


class DEPParser(NLPComponent):
    def init_state(self, graph: NLPGraph) -> NLPState:
        return DEPState(graph)





# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc

import numpy as np
from keras import backend as K
from keras.layers import Conv1D, Average, Multiply
from keras.layers import Dense, AveragePooling1D, Input, Lambda
from keras.models import Model
import tensorflow as tf

__author__ = 'Bonggun Shin, Jinho D. Choi'


class SentimentAnalyzer(object):
    def __init__(self, emb_model, model_path, maxlen=60):
        self.emb_model = emb_model
        self.model_path = model_path
        self.maxlen = maxlen
        self.p_model, self.a_model = self.load_model()
        self.graph = tf.get_default_graph()

    @abc.abstractmethod
    def prediction_model(self, model_input):
        return

    @abc.abstractmethod
    def attention_model(self, model_input):
        return

    def load_model(self):
        input_shape = (self.maxlen, self.emb_model.dim)
        model_input = Input(shape=input_shape)
        p_model = self.prediction_model(model_input)
        a_model = self.attention_model(model_input)

        for i in range(len(a_model.layers)):
            a_model.layers[i].set_weights(p_model.layers[i].get_weights())

        return p_model, a_model

    def decode(self, sentences, batch_size=2000, attn=False):
        x = self.emb_model.docs_to_emb(sentences, self.maxlen)
        with self.graph.as_default():
            y = self.p_model.predict(x, batch_size=batch_size, verbose=0)
            all_norm_att = []
            all_raw_att = []

            if not attn: return y, all_norm_att, all_raw_att
            sentence_len_list = [len(sentence) for sentence in sentences]
            attention_matrix = self.a_model.predict(x, batch_size=batch_size, verbose=0)

            for sample_index in range(len(sentence_len_list)):
                sample_norm_att_list = []
                sample_raw_att_list = []
                for gram_index in range(5):
                    sample_raw_att = attention_matrix[gram_index][sample_index][0]
                    sample_norm_att = sample_raw_att / max(sample_raw_att)
                    sample_norm_att_list.append(sample_norm_att[-sentence_len_list[sample_index] + gram_index:])
                    sample_raw_att_list.append(sample_raw_att[-sentence_len_list[sample_index] + gram_index:])

                new_norm_att = np.zeros([5, len(sample_norm_att_list[0])])
                new_norm_att[0, :] = sample_norm_att_list[0]

                new_raw_att = np.zeros([5, len(sample_raw_att_list[0])])
                new_raw_att[0, :] = sample_raw_att_list[0]

                raw_att_avg = np.zeros([1, len(sample_raw_att_list[0])])
                for i in range(1, 5):
                    for j in range(len(sample_norm_att_list[i])):
                        ngram = i + 1
                        for n in range(ngram):
                            new_norm_att[i, j + n] += sample_norm_att_list[i][j] / (ngram)
                            new_raw_att[i, j + n] += sample_raw_att_list[i][j] / (ngram)

                    new_norm_att[i] = new_norm_att[i] / max(new_norm_att[i])

                raw_att_avg[0] = new_raw_att.sum(0)
                norm_att_avg = raw_att_avg / max(raw_att_avg[0])
                all_norm_att.append(np.concatenate((norm_att_avg, new_norm_att), axis=0))
                all_raw_att.append(np.concatenate((raw_att_avg, new_raw_att), axis=0))

        return y, all_norm_att, all_raw_att


class TwitterSentimentAnalyzer(SentimentAnalyzer):
    def __init__(self, emb_model, model_path):
        super(TwitterSentimentAnalyzer, self).__init__(emb_model=emb_model, model_path=model_path, maxlen=60)

    def prediction_model(self, model_input):
        print('Init: '+self.model_path)
        filter_sizes = (1, 2, 3, 4, 5)
        num_filters = 80
        hidden_dims = 20
        conv_blocks = []
        for sz in filter_sizes:
            conv = Conv1D(num_filters,
                          sz,
                          padding="valid",
                          activation="relu",
                          strides=1)(model_input)
            print(conv)
            conv = Lambda(lambda x: K.permute_dimensions(x, (0, 2, 1)))(conv)
            conv = AveragePooling1D(pool_size=num_filters)(conv)

            attention_size = self.maxlen - sz + 1

            multiplied_vector_list = []
            for i in range(attention_size):
                selected_attention = Lambda(lambda x: x[:, 0, i] / float(sz))(conv)

                for j in range(sz):
                    selected_token = Lambda(lambda x: x[:, i + j, :])(model_input)
                    multiplied_vector = Lambda(lambda x: Multiply()(x))([selected_token, selected_attention])

                    multiplied_vector_list.append(multiplied_vector)

            attentioned_conv = Average()(multiplied_vector_list)

            print(attentioned_conv)
            conv_blocks.append(attentioned_conv)

        z = Average()(conv_blocks)
        z = Dense(hidden_dims, activation="relu")(z)
        model_output = Dense(3, activation="softmax")(z)

        model = Model(model_input, model_output)
        model.load_weights(self.model_path)
        model.compile(loss="sparse_categorical_crossentropy", optimizer="adam")

        return model

    def attention_model(self, model_input):
        filter_sizes = (1, 2, 3, 4, 5)
        num_filters = 80
        att_list = []
        for sz in filter_sizes:
            conv = Conv1D(num_filters,
                          sz,
                          padding="valid",
                          activation="relu",
                          strides=1)(model_input)
            conv = Lambda(lambda x: K.permute_dimensions(x, (0, 2, 1)))(conv)
            att = AveragePooling1D(pool_size=num_filters)(conv)
            att_list.append(att)

        model = Model(model_input, att_list)
        model.compile(loss="sparse_categorical_crossentropy", optimizer="adam")

        return model


class SSTSentimentAnalyzer(SentimentAnalyzer):
    def __init__(self, emb_model, model_path):
        super(SSTSentimentAnalyzer, self).__init__(emb_model=emb_model, model_path=model_path, maxlen=100)

    def prediction_model(self, model_input):
        print('Init: ' + self.model_path)
        filter_sizes = (1, 2, 3, 4, 5)
        num_filters = 64
        hidden_dims = 50
        conv_blocks = []
        for sz in filter_sizes:
            conv = Conv1D(num_filters,
                          sz,
                          padding="valid",
                          activation="relu",
                          strides=1)(model_input)
            print(conv)
            conv = Lambda(lambda x: K.permute_dimensions(x, (0, 2, 1)))(conv)
            conv = AveragePooling1D(pool_size=num_filters)(conv)

            attention_size = self.maxlen - sz + 1

            multiplied_vector_list = []
            for i in range(attention_size):
                selected_attention = Lambda(lambda x: x[:, 0, i] / float(sz))(conv)

                for j in range(sz):
                    selected_token = Lambda(lambda x: x[:, i + j, :])(model_input)
                    multiplied_vector = Lambda(lambda x: Multiply()(x))([selected_token, selected_attention])

                    multiplied_vector_list.append(multiplied_vector)

            attentioned_conv = Average()(multiplied_vector_list)

            print(attentioned_conv)
            conv_blocks.append(attentioned_conv)

        z = Average()(conv_blocks)
        z = Dense(hidden_dims, activation="relu")(z)
        model_output = Dense(3, activation="softmax")(z)

        model = Model(model_input, model_output)
        model.load_weights(self.model_path)
        model.compile(loss="sparse_categorical_crossentropy", optimizer="adam")

        return model

    def attention_model(self, model_input):
        filter_sizes = (1, 2, 3, 4, 5)
        num_filters = 64
        att_list = []
        for sz in filter_sizes:
            conv = Conv1D(num_filters,
                          sz,
                          padding="valid",
                          activation="relu",
                          strides=1)(model_input)
            conv = Lambda(lambda x: K.permute_dimensions(x, (0, 2, 1)))(conv)
            att = AveragePooling1D(pool_size=num_filters)(conv)
            att_list.append(att)

        model = Model(model_input, att_list)
        model.compile(loss="sparse_categorical_crossentropy", optimizer="adam")

        return model



# from elit.reader import TSVReader
# import unittest
# import os
#
# class ReaderTest(unittest.TestCase):
#     def setUp(self):
#         self.tsv_file = os.path.join(os.path.abspath(os.path.dirname(__file__)),
#                                      './../elit/resources/sample/sample.tsv')
#         self.reader = TSVReader(1, 2, 3, 4, 5, 6, 7, 8)
#
#     def test_tsv_reader_nodes(self):
#         self.reader.open(self.tsv_file)
#         self.assertEqual(len(self.reader.next_all), 2)
#         self.reader.close()
#
#     def test_tsv_reader_node(self):
#         self.reader.open(self.tsv_file)
#         nodes = [7, 11]
#         for i, node in enumerate(self.reader.next_all):
#             with self.subTest(i=i):
#                 self.assertEqual(len(node), nodes[i])
#         self.reader.close()
#
#     def test_tsv_reader_graph(self):
#         self.reader.open(self.tsv_file)
#         node = self.reader.next.nodes[1]
#         self.assertEqual(node.node_id, 1)
#         self.assertEqual(node.word, 'John')
#         self.assertEqual(node.lemma, 'john')
#         self.assertEqual(node.pos, 'NNP')
#         self.assertEqual(node.nament, 'U-PERSON')
#         self.assertEqual(node.feats, {})
#         self.reader.close()
#
#
# if __name__ == '__main__':
#     unittest.main()



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================


__author__ = 'T Lee'

#
# class StructureTest(unittest.TestCase):
#     def test_nlpnode_ancestor(self):
#         filename = 'resources/sample/sample.tsv'
#         reader = TSVReader(1, 2, 3, 4, 5, 6, 7, 8)
#         reader.open(filename)
#         graph = reader.next
#         nodes = []
#         for node in graph.nodes:
#             nodes.append(node)
#
#         self.assertEqual(nodes[0].parent, None)
#         self.assertEqual(nodes[0].grandparent, None)
#         self.assertEqual(nodes[3].parent, nodes[0])
#         self.assertEqual(nodes[8].parent, nodes[10])
#         self.assertEqual(nodes[8].grandparent, nodes[3])
#         self.assertTrue(nodes[3].parent_of(nodes[0]))
#         self.assertTrue(nodes[8].parent_of(nodes[10]))
#
#     def test_dependencyLabel(self):
#         filename = 'resources/sample/sample.tsv'
#         reader = TSVReader(1, 2, 3, 4, 5, 6, 7, 8)
#         reader.open(filename)
#         graph = reader.next
#         nodes = []
#         for node in graph.nodes:
#             nodes.append(node)
#
#         self.assertEqual(nodes[5].get_dependency_label(nodes[3]), 'advcl')
#         self.assertEqual(nodes[3].get_dependency_label(nodes[5]), None)
#         self.assertEqual(nodes[14].get_dependency_label(nodes[10]), 'ppmod')
#
#     def test_getChild(self):
#         filename = 'resources/sample/sample.tsv'
#         reader = TSVReader(1, 2, 3, 4, 5, 6, 7, 8)
#         reader.open(filename)
#         graph = reader.next
#         nodes = []
#         for node in graph.nodes:
#             nodes.append(node)
#
#         self.assertEqual(nodes[14].get_leftmost_child(), nodes[11])
#         self.assertEqual(nodes[19].get_leftmost_child(), nodes[16])
#         self.assertEqual(nodes[3].get_leftmost_child(), nodes[1])
#         self.assertEqual(nodes[3].get_rightmost_child(),nodes[20])
#         #self.assertEqual(nodes[19].get_rightmost_child(), nodes[18])
#         #self.assertEqual(nodes[0].get_leftmost_child(), nodes[3])
#
#     def test_nlpnode_set(self):
#         filename = 'resources/sample/sample.tsv'
#         reader = TSVReader(1, 2, 3, 4, 5, 6, 7, 8)
#         reader.open(filename)
#         graph = reader.next
#         container = []
#         for node in graph.nodes:
#             container.append(node)
#
#         tempnode1 = NLPNode("TEMP_1")
#         tempnode2 = NLPNode("TEMP_2")
#         container[0].set_parent(tempnode1, 'dep_1')
#         self.assertEqual(container[0].parent, tempnode1)
#         #container[0].set_parent(tempnode2, 'dep_2')
#         #self.assertNotEquals(container[0].parent, tempnode1)
#
#
# if __name__ == '__main__':
#     unittest.main()



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import re

from elit.dev.structure import DELIM_ARC, DELIM_ARC_KV, DELIM_FEAT_KV, BLANK, ROOT_TAG, \
    NLPToken, NLPGraph

__author__ = 'Jinho D. Choi'

_TAB = re.compile('\t')
_FEATS = re.compile('\\|')
_FEATS_KV = re.compile(DELIM_FEAT_KV)
_ARC = re.compile(DELIM_ARC)
_ARC_KV = re.compile(DELIM_ARC_KV)


class TSVReader:
    def __init__(self, word_index=-1, lemma_index=-1, pos_index=-1,
                 feats_index=-1, head_index=-1, deprel_index=-1,
                 sheads_index=-1, nament_index=-1):
        """

        :param word_index: the column index of word forms.1
        :type word_index: int
        :param lemma_index: the column index of lemma.
        :type lemma_index: int
        :param pos_index: the column index of part-of-speech tags.
        :type pos_index: int
        :param feats_index: the column index of extra features.
        :type feats_index: int
        :param head_index: the column index of primary head IDs.
        :type head_index: int
        :param deprel_index: the column index of primary dependency sys_labels.
        :type deprel_index: int
        :param sheads_index: the column index of secondary dependency heads.
        :type sheads_index: int
        :param nament_index: the column index of of named entity tags.
        :type nament_index: int
        """
        self.word_index = word_index
        self.lemma_index = lemma_index
        self.pos_index = pos_index
        self.feats_index = feats_index
        self.head_index = head_index
        self.deprel_index = deprel_index
        self.sheads_index = sheads_index
        self.nament_index = nament_index
        self.ins = None

    def __next__(self):
        graph = self.next
        if graph:
            return graph
        else:
            raise StopIteration

    def __iter__(self):
        return self

    @classmethod
    def create_reader(cls, reader):
        """

        :type reader: TSVReader
        :return: a reader adapting the configuration (not the input stream) from the other reader.
        :rtype: TSVReader
        """
        return TSVReader(word_index=reader.word_index,
                         lemma_index=reader.lemma_index,
                         pos_index=reader.pos_index,
                         feats_index=reader.feats_index,
                         head_index=reader.head_index,
                         deprel_index=reader.deprel_index,
                         sheads_index=reader.sheads_index,
                         nament_index=reader.nament_index)

    def open(self, filename: str):
        """

        :param filename: file
        :type filename: str
        :return: file
        :rtype: TextIOWrapper[str]
        """
        self.ins = open(filename)
        return self.ins

    def close(self):
        """
        Close file
        """
        self.ins.close()

    @property
    def next(self):
        """

        :return: get next graph
        :rtype: NLPGraph
        """
        tsv = []

        for line in self.ins:
            line = line.strip()
            if line:
                tsv.append(_TAB.split(line))
            elif tsv:
                break

        return self.tsv_to_graph(tsv) if tsv else None

    @property
    def next_all(self):
        """

        :return: List of NLPGraph
        :rtype: List[NLPGraph]
        """
        return [graph for graph in self]

    def tsv_to_graph(self, tsv):
        """

        :param tsv: each row represents a token, each column represents a field.
        :type tsv: List[List[str]]
        """

        def get_field(row, index):
            """

            :param row:
            :type row: List[str]
            :param index:
            :type index: int
            :return: ROOT_TAG
            :rtype: str
            """
            return ROOT_TAG if row is None else row[index] if index >= 0 else None

        def get_feats(row):
            """

            :param row:
            :type row: List[str]
            :return: feats
            :rtype: Union[Dict[str, str], None]
            """
            if self.feats_index >= 0:
                feats = row[self.feats_index]
                if feats == BLANK:
                    return None
                return {feat[0]: feat[1] for feat in map(_FEATS_KV.split, _FEATS.split(feats))}
            return None

        def init_node(i):
            """

            :param i:
            :type i: int
            :return: NLPNode
            :rtype: NLPToken
            """
            row = tsv[i]
            node_id = i + 1
            word = get_field(row, self.word_index)
            lemma = get_field(row, self.lemma_index)
            pos = get_field(row, self.pos_index)
            nament = get_field(row, self.nament_index)
            feats = get_feats(row) if row else None
            return NLPToken(token_id=node_id,
                            word=word,
                            lemma=lemma,
                            pos=pos,
                            nament=nament,
                            feats=feats)

        graph = NLPGraph([init_node(i) for i in range(len(tsv))])

        if self.head_index >= 0:
            for i, node in enumerate(graph):
                _tsv = tsv[i]
                node.set_parent(graph.nodes[int(_tsv[self.head_index])], _tsv[self.deprel_index])

                if self.sheads_index >= 0 and _tsv[self.sheads_index] != BLANK:
                    for arc in map(_ARC_KV.split, _ARC.split(_tsv[self.sheads_index])):
                        node.add_secondary_parent(graph.nodes[int(arc[0])], arc[1])

        return graph




# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
from elit.lexicon import Word2Vec
from elit.sentiment.sentiment_analyzer import TwitterSentimentAnalyzer, SSTSentimentAnalyzer
from elit.tokenizer import english_tokenizer

__author__ = 'Bonggun Shin'

# os.environ["CUDA_VISIBLE_DEVICES"] = '0'
EMB_FILE = '/Users/jdchoi/Downloads/tmp/w2v/w2v-400-twitter.gnsm'
MODEL_PATH = '/Users/jdchoi/Downloads/tmp/model/sentiment-semeval17-400-v2'

emb_model = Word2Vec(EMB_FILE)
sa = TwitterSentimentAnalyzer(emb_model, MODEL_PATH)

sentences = ["I feel a little bit tired today, but I am really happy!",
             "Although the rain stopped, I hate this thick cloud in the sky."]

tokenized_sentences = [[token[0] for token in english_tokenizer.tokenize(s, False)] for s in sentences]

y, att, raw_att = sa.decode(tokenized_sentences, attn=True)
print(y)
print(att)
print(raw_att)

EMB_FILE = '/Users/jdchoi/Downloads/tmp/w2v/w2v-400-amazon-review.gnsm'
MODEL_PATH = '/Users/jdchoi/Downloads/tmp/model/sentiment-sst-400-v2'

emb_model = Word2Vec(EMB_FILE)
sa = SSTSentimentAnalyzer(emb_model, MODEL_PATH)

sentences = ["This is a film well worth seeing, talking and singing heads and all.",
             "The sort of movie that gives tastelessness a bad rap.",
             "Marisa Tomei is good, but Just A Kiss is just a mess."]

tokenized_sentences = [[token[0] for token in english_tokenizer.tokenize(s, False)] for s in sentences]

y, att, raw_att = sa.decode(tokenized_sentences, attn=True)
print(y)
print(att)
print(raw_att)


# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import argparse
import logging
import os
from typing import List, Tuple, Callable

import mxnet as mx
from elit.structure import NLPGraph

from elit.dev.reader import TSVReader

__author__ = 'Jinho D. Choi'


# ============================== Neural Networks ==============================

def create_ffnn(hidden: List[Tuple[int, str, float]], input_dropout: float=0, output_size: int=2,
                context: mx.context.Context=mx.cpu()) -> mx.mod.Module:
    net = mx.sym.Variable('data')
    if input_dropout > 0: net = mx.sym.Dropout(net, p=input_dropout)

    for i, (num_hidden, act_type, dropout) in enumerate(hidden, 1):
        net = mx.sym.FullyConnected(net, num_hidden=num_hidden, name='fc'+str(i))
        if act_type: net = mx.sym.Activation(net, act_type=act_type, name=act_type+str(i))
        if dropout > 0: net = mx.sym.Dropout(net, p=dropout)

    net = mx.sym.FullyConnected(net, num_hidden=output_size, name='fc' + str(len(hidden) + 1))
    net = mx.sym.SoftmaxOutput(net, name='softmax')

    return mx.mod.Module(symbol=net, context=context)


def conv_pool(net: mx.sym.Variable, conv_kernel: Tuple[int, int], num_filter: int, act_type: str,
              pool_kernel: Tuple[int, int], pool_type='max', pool_stride: Tuple[int, int]=(1, 1)) -> mx.sym.Variable:
    net = mx.sym.Convolution(data=net, kernel=conv_kernel, num_filter=num_filter)
    net = mx.sym.Activation(data=net, act_type=act_type)
    net = mx.sym.Pooling(data=net, pool_type=pool_type, kernel=pool_kernel, stride=pool_stride)
    return net

# ============================== Reader ==============================

def read_graphs(reader: TSVReader, filename: str) -> List[NLPGraph]:
    logging.info('Reading: '+os.path.basename(filename))
    reader.open(filename)
    graphs = reader.next_all
    reader.close()
    logging.info('- %s graphs' % len(graphs))
    return graphs


# ============================== Argument ==============================

def argparse_data(parser: argparse.ArgumentParser, tsv: Callable[[Tuple[int]], TSVReader]=None):
    """
    :param parser: the parent parser.
    :param tsv: (indices -> TSVReader, comment).
    """
    args = parser.add_argument_group('Data')

    args.add_argument('--trn_data', type=str, metavar='filepath', help='path to the training data')
    args.add_argument('--dev_data', type=str, metavar='filepath', help='path to the development data')

    if tsv:
        def reader(s: str):
            t = tuple(map(int, s.split(',')))
            return tsv(t)

        args.add_argument('--tsv', type=reader, metavar='int(,int)*', help='indices for the TSV reader')

    return args


def argparse_lexicon(parser: argparse.ArgumentParser):
    args = parser.add_argument_group('Lexicon')

    args.add_argument('--w2v', type=str, metavar='filepath', help='path to the word2vec bin file')
    args.add_argument('--f2v', type=str, metavar='filepath', help='path to the fasttext bin file')

    return args


def argparse_model(parser: argparse.ArgumentParser):
    def context(s: str):
        m = mx.cpu if s[0] == 'c' else mx.gpu
        s = s[1:]

        if ',' in s:
            r = tuple(map(int, s.split(',')))
        elif '-' in s:
            t = tuple(map(int, s.split('-')))
            r = range(t[0], t[1]+1)
        else:
            r = [int(s)]

        return m(r[0]) if len(r) == 1 else [m(i) for i in r]

    model = parser.add_argument_group('Model')

    model.add_argument('--num_steps', type=int, metavar='int', default=1000,
                       help='number of steps for training')
    model.add_argument('--batch_size', type=int, metavar='int', default=128,
                       help='size of the mini batch')
    model.add_argument('--bagging_ratio', type=float, metavar='float', default=0.63,
                       help='ratio for the bootstrap aggregating')
    model.add_argument('--context', type=context, metavar='g|c:int(,int)*|int-int', default=mx.cpu(),
                       help='context used for the module')
    model.add_argument('--optimizer', type=str, metavar='sgd|adagrad|adam', default='sgd',
                       help='optimizer for training')

    return model


def argparse_ffnn(parser: argparse.ArgumentParser):
    def hidden(s: str):
        size, activation, dropout = s.split(',')
        return int(size), activation, float(dropout)

    ffnn = parser.add_argument_group('Model', 'Feed forward neural network')

    ffnn.add_argument('--hidden', nargs='*', type=hidden, metavar='int,str,float', default=[],
                      help='(size, act type, dropout rate) for hidden layers')
    ffnn.add_argument('--input_dropout', type=float, metavar='float', default=0,
                      help='dropout rate for input layer')
    ffnn.add_argument('--output_size', type=int, metavar='int', default=2,
                      help='size of the output layer')


# def add_arguments_fit(parser: argparse.ArgumentParser):
#     def context(s: str):
#         try:
#
#         except:
#             raise argparse.ArgumentTypeError('Invalid format: '+s)
#
#
#     parser.add_argument('--context', nargs='+' type=str,
#                        help='cpu|gpu ')
#     train.add_argument('--gpus', type=str,
#                        help='list of gpus to run, e.g. 0 or 0,2,5. empty means using cpu')
#     train.add_argument('--kv-store', type=str, zero='device',
#                        help='key-value store type')
#     train.add_argument('--num-epochs', type=int, zero=100,
#                        help='max num of epochs')
#     train.add_argument('--lr', type=float, zero=0.1,
#                        help='initial learning rate')
#     train.add_argument('--lr-factor', type=float, zero=0.1,
#                        help='the ratio to reduce lr on each step')
#     train.add_argument('--lr-step-epochs', type=str,
#                        help='the epochs to reduce the lr, e.g. 30,60')
#     train.add_argument('--optimizer', type=str, zero='sgd',
#                        help='the optimizer type')
#     train.add_argument('--mom', type=float, zero=0.9,
#                        help='momentum for sgd')
#     train.add_argument('--wd', type=float, zero=0.0001,
#                        help='weight decay for sgd')
#     train.add_argument('--batch-size', type=int, zero=128,
#                        help='the batch size')
#     train.add_argument('--disp-batches', type=int, zero=20,
#                        help='show progress for every n batches')
#     train.add_argument('--model-prefix', type=str,
#                        help='model prefix')
#     parser.add_argument('--monitor', dest='monitor', type=int, zero=0,
#                         help='log network parameters every N iters if larger than 0')
#     train.add_argument('--load-epoch', type=int,
#                        help='load the model on an epoch using the model-load-prefix')
#     train.add_argument('--top-k', type=int, zero=0,
#                        help='report the top-k accuracy. 0 means no report.')
#     train.add_argument('--test-io', type=int, zero=0,
#                        help='1 means test reading speed without training')
#     return train


# def create_argparser(description: str) -> argparse.ArgumentParser:
#
#     parser = argparse.ArgumentParser(description=description)
#
#     # neural networks
#     add_arguments_ffnn(parser)
#
#
#     # data
#     parser.add_argument('--train', nargs=1, type=str, help='path to the training data')
#     parser.add_argument('--develop', nargs=1, type=str, help='path to the development data')
#
#     # mxnet
#     parser.add_argument('--context', nargs=2, type)
#
#     return parser



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
from abc import ABCMeta
from abc import abstractmethod
from typing import Union, List

import numpy as np

from elit.dev.structure import NLPGraph, NLPToken, Relation

__author__ = 'Jinho D. Choi'


class NLPState(metaclass=ABCMeta):
    def __init__(self, graph: NLPGraph):
        self.graph: NLPGraph = graph
        self.reset_count = 0

    @abstractmethod
    def reset(self):
        """ Reset to the initial state and remove all tags. """

    # ============================== Oracle ==============================

    @property
    @abstractmethod
    def gold(self) -> str:
        """
        :return: the gold label for the current state if exists; otherwise, None.
        """

    @abstractmethod
    def eval(self, stats: np.array) -> float:
        """
        :param stats: accumulated statistics for evaluation.
        :return: the overall accuracy based on the statistics.
        """

    # ============================== Transition ==============================

    @abstractmethod
    def process(self, label: str, scores: np.array=None):
        """
        :param label: the label given the current state.
        :param scores: prediction scores.
          Apply the label to the current state and move onto the next state.
        """

    @abstractmethod
    def terminate(self) -> bool:
        """
        :return: True if no more state can be processed; otherwise, False.
        """

    # ============================== Feature ==============================

    @abstractmethod
    def features(self, node: NLPToken) -> List[np.array]:
        """
        :return: features given the current state.
        """

    def get_node(self, index: int, window: int=0, relation: Relation=None, root: bool=False) -> Union[NLPToken, None]:
        """
        :param index: the index of the anchor node.
        :param window: the context window to the anchor node.
        :param relation: the relation to the (index+window)'th node.
        :param root: if True, the root (nodes[0]) is returned when the condition is met; otherwise, None.
        :return: the relation(index+window)'th node in the graph if exists; otherwise, None.
        """
        index += window
        begin = 0 if root else 1
        node: NLPToken = self.graph.nodes[index] if begin <= index < len(self.graph) else None

        if node and relation:
            # 1st order
            if relation == Relation.PARENT:
                return node.parent
            if relation == Relation.LEFTMOST_CHILD:
                return node.get_leftmost_child()
            if relation == Relation.RIGHTMOST_CHILD:
                return node.get_rightmost_child()
            if relation == Relation.LEFT_NEAREST_CHILD:
                return node.get_left_nearest_child()
            if relation == Relation.RIGHT_NEAREST_CHILD:
                return node.get_right_nearest_child()
            if relation == Relation.LEFTMOST_SIBLING:
                return node.get_leftmost_sibling()
            if relation == Relation.RIGHTMOST_SIBLING:
                return node.get_rightmost_sibling()
            if relation == Relation.LEFT_NEAREST_SIBLING:
                return node.get_left_nearest_sibling()
            if relation == Relation.RIGHT_NEAREST_SIBLING:
                return node.get_right_nearest_sibling()

            # 2nd order
            if relation == Relation.GRANDPARENT:
                return node.grandparent
            if relation == Relation.SND_LEFTMOST_CHILD:
                return node.get_leftmost_child(1)
            if relation == Relation.SND_RIGHTMOST_CHILD:
                return node.get_rightmost_child(1)
            if relation == Relation.SND_LEFT_NEAREST_CHILD:
                return node.get_left_nearest_child(1)
            if relation == Relation.SND_RIGHT_NEAREST_CHILD:
                return node.get_right_nearest_child(1)
            if relation == Relation.SND_LEFTMOST_SIBLING:
                return node.get_leftmost_sibling(1)
            if relation == Relation.SND_RIGHTMOST_SIBLING:
                return node.get_rightmost_sibling(1)
            if relation == Relation.SND_LEFT_NEAREST_SIBLING:
                return node.get_left_nearest_sibling(1)
            if relation == Relation.SND_RIGHT_NEAREST_SIBLING:
                return node.get_right_nearest_sibling(1)

        return node

    def is_first(self, node: NLPToken) -> bool:
        """
        :param node: the node to be compared
        :return: True if the node is the first node in the graph; otherwise, False
        """
        return len(self.graph) > 1 and self.graph.nodes[1] == node

    def is_last(self, node: NLPToken) -> bool:
        """
        :param node: the node to be compared
        :return: True if the node is the last node in the graph; otherwise, False
        """
        return self.graph.nodes[-1] == node



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================

from elit.api import ELITDecoder
from io import StringIO

__author__ = 'Jinho D. Choi'

flag = '0114'
input_text = 'I watched "The Sound of Music" last night. It is my favorite movie.'
nd = ELITDecoder(resource_dir='../../resources/', sentiment_twitter=False)

# returns the output as a list of documents
docs = nd.decode(flag, StringIO(input_text))
print(docs)

# saves the output as a JSON file: out.json
fout = open('out.json', 'w')
nd.decode(flag, StringIO(input_text), fout)




# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================

import sys
sys.path.append("/Users/tlee/Desktop/elit")

import argparse
import logging
from typing import Tuple, List

import mxnet as mx
import numpy as np
import fasttext
from fasttext.model import WordVectorModel
from gensim.models import KeyedVectors

from elit.dev.template.lexicon import NLPLexiconMapper, NLPEmbedding
from elit.dev.template.model import NLPModel
from elit.dev.template.state import NLPState
from elit.dev.template.util import argparse_ffnn, argparse_model, argparse_data, read_graphs, argparse_lexicon, conv_pool
from elit.dev.reader import TSVReader
from elit.dev.structure import NLPGraph, NLPToken

__author__ = 'Jinho D. Choi'


class POSLexicon(NLPLexiconMapper):
    def __init__(self, w2v: KeyedVectors=None, f2v: WordVectorModel=None, a2v: KeyedVectors=None,
                 output_size: int=50):
        """
        :param w2v: word embeddings from word2vec.
        :param f2v: word embeddings from fasttext.
        :param a2v: a2v classes.
        :param output_size: the number of part-of-speech tags to predict.
        """
        super().__init__(w2v, f2v)
        self.a2v: NLPEmbedding = NLPEmbedding(a2v, 'word', 'a2v') if a2v else None
        self.pos_zeros = np.zeros((output_size,)).astype('float32')


class POSState(NLPState):
    def __init__(self, graph: NLPGraph, lexicon: POSLexicon, save_gold=False):
        super().__init__(graph)
        self.lex: POSLexicon = lexicon

        # reset
        self.golds = [node.set_pos(None) for node in self.graph] if save_gold else None
        for node in self.graph: node.pos_scores = lexicon.pos_zeros
        self.idx_curr: int = 1

    def reset(self):
        for node in self.graph:
            node.pos = None
            node.pos_scores = self.lex.pos_zeros

        self.idx_curr = 1
        self.reset_count += 1

    # ============================== Oracle ==============================

    @property
    def gold(self) -> str:
        return self.golds[self.idx_curr - 1] if self.golds else None

    def eval(self, stats: np.array) -> float:
        if self.golds is None: return 0

        stats[0] += len(self.graph)
        for i, node in enumerate(self.graph):
            if node.pos == self.golds[i]:
                stats[1] += 1

        return stats[1] / stats[0]

    # ============================== Transition ==============================

    def process(self, label: str, scores: np.array=None):
        node: NLPToken = self.graph.nodes[self.idx_curr]
        if scores is not None: node.pos_scores = scores
        node.pos = label
        self.idx_curr += 1

    @property
    def terminate(self) -> bool:
        return self.idx_curr >= len(self.graph.nodes)

    # ============================== Feature ==============================

    def features(self, node: NLPToken) -> List[np.array]:
        fs = [node.pos_scores if node else self.lex.pos_zeros]
        if self.lex.w2v: fs.append(self.lex.w2v.get(node))
        if self.lex.f2v: fs.append(self.lex.f2v.get(node))
        if self.lex.a2v: fs.append(self.lex.a2v.get(node))
        return fs


class POSModel(NLPModel):
    def __init__(self, batch_size=64, num_label: int=50, feature_context: Tuple = (-2, -1, 0, 1, 2),
                 context: mx.context.Context=mx.cpu(0), w2v_dim=100,
                 ngram_filter_list=(1, 2, 3), ngram_filter: int=32):
        super().__init__(POSState, batch_size)
        self.mxmod: mx.module.Module = self.init_mxmod(batch_size=batch_size,
                                                       num_label=num_label,
                                                       num_feature=len(feature_context),
                                                       context=context,
                                                       w2v_dim=w2v_dim,
                                                       ngram_filter_list=ngram_filter_list,
                                                       ngram_filter=ngram_filter)
        self.feature_context: Tuple[int] = feature_context

    # ============================== Feature ==============================

    def x(self, state: POSState) -> np.array:
        vectors_pos_score = [state.features(state.get_node(state.idx_curr, window))[0] for window in self.feature_context]
        vectors_f2v = [state.features(state.get_node(state.idx_curr, window))[1] for window in self.feature_context]
        vectors_a2v = [state.features(state.get_node(state.idx_curr, window))[2] for window in self.feature_context]

        out_f2v = np.asarray(vectors_f2v)
        out_a2v = np.asarray(vectors_a2v)
        out = (out_f2v, out_a2v)
        return out

    # ============================== Module ==============================

    def init_mxmod(self, batch_size: int, num_label: int, num_feature: int, context: mx.context.Context, w2v_dim: int,
                   ngram_filter_list: Tuple, ngram_filter: int) -> mx.module.Module:
        batch_size = -1
        # n-gram convolution for f2v and a2v
        input_f2v = mx.sym.Variable('data_f2v')
        input_a2v = mx.sym.Variable('data_a2v')
        input_pool2v = mx.sym.Variable('data_pool2v')

        print ("NUMBER FEATURE: ", num_feature)
        conv_input_f2v = mx.sym.Reshape(data=input_f2v, shape=(batch_size, 1, num_feature, w2v_dim))
        conv_input_a2v = mx.sym.Reshape(data=input_a2v, shape=(batch_size, 1, num_feature, 50))
        conv_input_pool2v = mx.sym.Reshape(data=input_pool2v, shape=(batch_size, 1, 1, 192))


        pooled_1 = [conv_pool(conv_input_f2v, conv_kernel=(filter, w2v_dim), num_filter=ngram_filter, act_type='relu',
                            pool_kernel=(num_feature - filter + 1, 1), pool_stride=(1, 1))
                  for filter in ngram_filter_list]

        pooled_2 = [conv_pool(conv_input_a2v, conv_kernel=(filter, 50), num_filter=ngram_filter, act_type='relu',
                            pool_kernel=(num_feature - filter + 1, 1), pool_stride=(1, 1))
                  for filter in ngram_filter_list]


        # concatenate pooled features from f2v and a2v
        pooled = pooled_1 + pooled_2
        concat = mx.sym.Concat(*pooled, dim=1)

        h_pool = mx.sym.Reshape(name="concat_pooling", data=concat, shape=(batch_size, 2 * ngram_filter * len(ngram_filter_list)))
      # h_pool = mx.sym.Dropout(data=h_pool, p=dropouts[0]) if dropouts[0] > 0.0 else h_pool

        # block gradient
        h_pool = mx.sym.BlockGrad(h_pool, name="concat_pooling")

        # fully connected
        fc_weight = mx.sym.Variable('fc_weight')
        fc_bias = mx.sym.Variable('fc_bias')
        fc = mx.sym.FullyConnected(data=h_pool, weight=fc_weight, bias=fc_bias, num_hidden=num_label)

        output = mx.sym.Variable('softmax_label')
        sm = mx.sym.SoftmaxOutput(data=fc, label=output, name='softmax')

        # mx module now contains softmax and pool output
        final = mx.sym.Group([sm, h_pool])
        return mx.mod.Module(symbol=final, data_names=('data_f2v', 'data_a2v'), context=context)


def parse_args():
    parser = argparse.ArgumentParser('Train a part-of-speech tagger')

    # data
    args= argparse_data(parser, tsv=lambda t: TSVReader(word_index=t[0], pos_index=t[1]))
    args.add_argument('--log', type=str, metavar='filepath', help='path to the logging file')

    # lexicon
    args = argparse_lexicon(parser)
    args.add_argument('--a2v', type=str, metavar='filepath', help='path to the ambiguity class bin file')

    # model
    def feature_context(s: str):
        return tuple(map(int, s.split(',')))

    argparse_ffnn(parser)
    args = argparse_model(parser)
    args.add_argument('--feature_context', type=feature_context, metavar='int,int*', default=[-2, -1, 0, 1, 2],
                      help='context window for feature extraction')

    return parser.parse_args()


def main():
    # arguments
    args = parse_args()
    if args.log: logging.basicConfig(filename=args.log, format='%(message)s', level=logging.INFO)
    else: logging.basicConfig(format='%(message)s', level=logging.INFO)

    # data
    trn_graphs = read_graphs(args.tsv, args.trn_data)
    dev_graphs = read_graphs(args.tsv, args.dev_data)

    # lexicon
    w2v = KeyedVectors.load_word2vec_format(args.w2v, binary=True) if args.w2v else None
    f2v = fasttext.load_model(args.f2v) if args.f2v else None
    a2v = KeyedVectors.load_word2vec_format(args.a2v, binary=True) if args.a2v else None

    lexicon = POSLexicon(w2v=w2v, f2v=f2v, a2v=a2v, output_size=args.output_size)

    # model
    model = POSModel(feature_context=args.feature_context, batch_size=64, w2v_dim=100)
    model.train(trn_graphs, dev_graphs, lexicon, num_steps=args.num_steps,
                bagging_ratio=args.bagging_ratio, optimizer=args.optimizer, force_init=True)


if __name__ == '__main__':
    main()



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import mxnet as mx
import time
import sys

__author__ = 'Jinho D. Choi'

vocab_size = 40000
batch_size = 50
emb_dim = 200
document_size = 256

input_x = mx.sym.Variable('data')
input_y = mx.sym.Variable('softmax_label')
emb_layer = mx.sym.Embedding(data=input_x, input_dim=vocab_size, output_dim=emb_dim, name='vocab_emb')
conv_input = mx.sym.Reshape(data=emb_layer, target_shape=(batch_size, 1, document_size, emb_dim))

filter_sizes = [2, 3, 4, 5]
num_filter = 100
pooled_outputs = []

for i, filter_size in enumerate(filter_sizes):
    convi = mx.sym.Convolution(data=conv_input, kernel=(filter_size, emb_dim), num_filter=num_filter)
    relui = mx.sym.Activation(data=convi, act_type='relu')
    pooli = mx.sym.Pooling(data=relui, pool_type='max', kernel=(document_size - filter_size + 1, 1), stride=(1, 1))
    pooled_outputs.append(pooli)

total_filters = num_filter * len(filter_sizes)
concat = mx.sym.Concat(*pooled_outputs, dim=1)
h_pool = mx.sym.Reshape(data=concat, target_shape=(batch_size, total_filters))

dropout = 0.5
h_drop = mx.sym.Dropout(data=h_pool, p=dropout) if dropout > 0 else h_pool

num_labels = 2
cls_weight = mx.sym.Variable('cls_weight')
cls_bias = mx.sym.Variable('cls_bias')
fc = mx.sym.FullyConnected(data=h_drop, weight=cls_weight, bias=cls_bias, num_hidden=num_labels)
sm = mx.sym.SoftmaxOutput(data=fc, label=input_y, name='softmax')
cnn = sm


epochs = 10
learning_rate = 0.01
opt = mx.optimizer.create('rmsprop')
opt.lr = learning_rate
updater = mx.optimizer.get_updater(opt)
logs = sys.stderr

for iter in range(epochs):
    tic = time.time()
    correct = 0
    total = 0



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import numpy as np
from gensim.models import KeyedVectors

__author__ = 'Jinho D. Choi'


class Word2Vec:
    def __init__(self, filepath):
        """
        :param filepath: the path to the file containing word embeddings.
        """
        self.model = None

        if filepath.endswith('.gnsm'):
            self.model = KeyedVectors.load(filepath)
        elif filepath.endswith('.bin'):
            self.model = KeyedVectors.load_word2vec_format(filepath, binary=True)
        else:
            raise ValueError('Unknown type: ' + filepath)

        self.dim = self.model.syn0.shape[1]
        self.pad = np.zeros((self.dim,)).astype('float32')
        print('Init: %s (vocab = %d, dim = %d)' % (filepath, len(self.model.vocab), self.dim))

    def doc_to_emb(self, document, maxlen):
        """
        :param document: the document comprising tokens to retrieve word embeddings for.
        :param maxlen: the maximum length of the document (# of tokens).
        :return: the list of word embeddings corresponding to the tokens in the document.
        """
        def emb(token_index):
            if token_index >= len(document): return self.pad
            v = self.model.vocab.get(document[token_index], None)
            return self.model.syn0[v.index] if v is not None else self.pad

        # return np.array([emb(i) for i in range(maxlen)])
        # TODO: the following 3 lines should be replaced by the above return statement
        l = [self.model.syn0[0] for _ in range(maxlen-len(document))]
        l.extend([emb(i) for i in range(min(maxlen, len(document)))])
        return np.array(l)

    def docs_to_emb(self, documents, maxlen):
        """
        :param documents: a list of documents.
        :param maxlen:
        :return:
        """
        return np.array([self.doc_to_emb(document, maxlen) for document in documents])


# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================


# import unittest
#
# from elit.dev.template.model import NLPModel
#
# __author__ = 'T Lee'
#
#
# class ModelTest(unittest.TestCase):
#     # TODO:
#     def setUp(self):
#         pass
#
#
# if __name__ == '__main__':
#     unittest.main()



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import functools
from enum import Enum
from itertools import islice

from elit.dev.util.bisect import bisect_index
from elit.dev.util.bisect import bisect_left
from elit.dev.util.bisect import bisect_remove
from elit.dev.util.bisect import bisect_right
from elit.dev.util.bisect import insort_right

__author__ = 'Jinho D. Choi'

# fields
BLANK = '_'
ROOT_TAG = '@#r$%'

# delimiters
DELIM_FEAT = '|'
DELIM_FEAT_KV = '='
DELIM_ARC = ';'
DELIM_ARC_KV = ':'


@functools.total_ordering
class NLPToken:
    def __init__(self, sen_id=-1, token_id=-1, form=None, lemma=None, pos=None, nament=None, feats=None):
        """
        :param sen_id: sentence id.
        :type sen_id: int
        :param token_id: node id.
        :type token_id: int
        :param form: word form.
        :type form: str
        :param lemma: lemma.
        :type lemma: str
        :param pos: part-of-speech tag.
        :type pos: str
        :param nament: named entity tag.
        :type nament: str
        :param feats: extra features.
        :type feats: Dict[str, str]
        """
        # fields
        self.sen_id = sen_id
        self.token_id = token_id
        self.form = form
        self.lemma = lemma
        self.pos = pos
        self.nament = nament
        self.feats = feats or {}

        # dependencies
        self.parent = None
        self.children = []
        self.secondary_parents = []
        self.secondary_children = []
        self.deprels = {}

    def __hash__(self):
        return hash(id(self))

    def __lt__(self, other):
        return self.token_id < other.node_id

    def __eq__(self, other):
        return id(self) == id(other)

    def __str__(self):
        node_id = str(self.token_id)
        form = self.form if self.form else BLANK
        lemma = self.lemma if self.lemma else BLANK
        pos = self.pos if self.pos else BLANK
        nament = self.nament if self.nament else BLANK
        feats = DELIM_FEAT.join((DELIM_FEAT_KV.join((k, v)) for k, v in self.feats.items())) if self.feats else BLANK
        head_id = str(self.parent.node_id) if self.parent else BLANK
        deprel = self.get_dependency_label(self.parent) or BLANK
        sheads = DELIM_ARC.join(DELIM_ARC_KV.join((str(parent.node_id), self.get_dependency_label(parent)))
                                for parent in self.secondary_parents) if self.secondary_parents else BLANK
        return '\t'.join((node_id, form, lemma, pos, feats, head_id, deprel, sheads, nament))

    def set_pos(self, pos):
        """

        :param pos: the part-of-speech tag to be assigned to this node.
        :type pos: str
        :return: the previous part-of-speech tag if exists; otherwise, None.
        :rtype: str
        """
        self.pos, prev = pos, self.pos
        return prev

    @classmethod
    def root(cls):
        """

        :return: cls
        :rtype: cls
        """
        return cls(node_id=0, word=ROOT_TAG, lemma=ROOT_TAG, pos=ROOT_TAG, nament=ROOT_TAG)

    @property
    def grandparent(self):
        """

        :return: the GRANDPARENT of this node if exists; otherwise, None.
        :rtype: NLPToken
        """
        return self.parent.parent if self.parent else None

    def get_dependency_label(self, node=None):
        """
        :param node: the parent of this node.
        :type node: NLPToken
        :return: the dependency label between this node and the PARENT node
                if exists; otherwise, None.
        :rtype: str
        """
        if node is None:
            node = self.parent
        return self.deprels.get(node, None) if node else None

    def set_dependency_label(self, node, label):
        """
        :param node: the PARENT of this node.
        :type node: NLPToken
        :param label: the dependency relation to the PARENT.
        :type label: str
        """
        if label:
            self.deprels[node] = label

    def set_parent(self, node, label=None):
        """
        :param node: the node to be set as the PARENT of this node.
        :type node: NLPToken
        :param label: the dependency relation between this node and the PARENT.
        :type label: str
        :return the previous PARENT if exists; otherwise, None.
        :rtype: NLPToken
        """
        # handle the previous PARENT
        prev_parent = self.parent

        if prev_parent:
            bisect_remove(prev_parent.children, self)
            del self.deprels[prev_parent]

        # set the current PARENT
        self.parent = node

        if node:
            insort_right(node.children, self)
            self.set_dependency_label(node, label)

        return prev_parent

    def child_of(self, node):
        """
        :param node: the node to be compared.
        :type node: NLPToken
        :return: True if this node is a child of the specific node; otherwise, False.
        :rtype: bool
        """
        return self.parent and self.parent == node

    def add_secondary_parent(self, node, label=None):
        """
        :param node: the node to be added as a secondary PARENT.
        :type node: NLPToken
        :param label: the dependency relation to the PARENT.
        :type label: str
        """
        insort_right(self.secondary_parents, node)
        insort_right(node.secondary_children, self)
        self.set_dependency_label(node, label)

    def remove_secondary_parent(self, node):
        """
        :param node: the node to be removed from the secondary PARENT list.
        :type node: NLPToken
        :return: True if the node is removed successfully; otherwise, False.
        :rtype: bool
        """
        idx = bisect_index(self.secondary_parents, node)
        if idx >= 0:
            del self.secondary_parents[idx]
            self.deprels.pop(node, None)
            return True
        return False

    def get_leftmost_child(self, order=0):
        """
        :param order: order displacement (0: leftmost, 1: 2nd leftmost, etc.).
        :type order: int
        :return: the leftmost child whose token position is on the left-hand side of this node
                if exists; otherwise, None.
        :rtype: Union['NLPNode', None]
        """
        idx = order
        return self.children[idx] if 0 <= idx < len(self.children) and self.children[idx] < self else None

    def get_rightmost_child(self, order=0):
        """
        :param order: order displacement (0: rightmost, 1: 2nd rightmost, etc.).
        :type order: int
        :return: the rightmost child whose token position is on the right-hand side of this node
                if exists; otherwise, None.
        :rtype: Union['NLPNode', None]
        """
        idx = len(self.children) - 1 - order
        return self.children[idx] if 0 <= idx < len(self.children) and self.children[idx] > self else None

    def get_left_nearest_child(self, order=0):
        """
        :param order: order displacement (0: left-nearest, 1: 2nd left-nearest, etc.).
        :type order: int
        :return: the left-nearest child whose token position is on the left-hand side of this node
                if exists; otherwise, None.
        :rtype: Union['NLPNode', None]
        """
        idx = bisect_left(self.children, self) - 1 - order
        return self.children[idx] if 0 <= idx < len(self.children) else None

    def get_right_nearest_child(self, order=0):
        """
        :param order: order displacement (0: right-nearest, 1: 2nd right-nearest, etc.).
        :type order: int
        :return: the right-nearest primary child whose token position is on the right-hand side of
                this node if exists; otherwise, None.
        :rtype: Union['NLPNode', None]
        """
        idx = bisect_right(self.children, self) + order
        return self.children[idx] if 0 <= idx < len(self.children) else None

    def get_leftmost_sibling(self, order=0):
        """
        :param order: order displacement (0: leftmost, 1: 2nd leftmost, etc.).
        :type order: int
        :return: the leftmost primary sibling whose token position is on the left-hand side of
                this node if exists; otherwise, None.
        :rtype: Union['NLPNode', None]
        """
        idx = order
        return self.parent.children[idx] \
            if self.parent and 0 <= idx < len(self.parent.children) and self.parent.children[idx] < self else None

    def get_rightmost_sibling(self, order=0):
        """
        :param order: order displacement (0: rightmost, 1: 2nd rightmost, etc.).
        :type order: int
        :return: the rightmost primary sibling whose token position is on the right-hand side of
                this node if exists; otherwise, None.
        :rtype: Union['NLPNode', None]
        """
        idx = len(self.children) - 1 - order
        return self.parent.children[idx] \
            if self.parent and 0 <= idx < len(self.parent.children) and self.parent.children[idx] > self else None

    def get_left_nearest_sibling(self, order=0):
        """
        :param order: order displacement (0: left-nearest, 1: 2nd left-nearest, etc.).
        :type order: int
        :return: the left-nearest primary sibling whose token position is on the left-hand side of
                this node if exists; otherwise, None.
        :rtype: Union['NLPNode', None]
        """
        if self.parent:
            idx = bisect_left(self.parent.children, self) - 1 - order
            return self.parent.children[idx] if 0 <= idx < len(self.parent.children) else None
        return None

    def get_right_nearest_sibling(self, order=0):
        """
        :param order: order displacement (0: right-nearest, 1: 2nd right-nearest, etc.).
        :type order: int
        :return: the right-nearest primary sibling whose token position is on the right-hand side of
                this node if exists; otherwise, None.
        :rtype: Union['NLPNode', None]
        """
        if self.parent:
            idx = bisect_right(self.parent.children, self) + 1 + order
            return self.parent.children[idx] if 0 <= idx < len(self.parent.children) else None
        return None


class NLPGraph:

    def __init__(self, nodes=None):
        """
        :param nodes: a list of NLP nodes whose parents are not initialized.
        :type  nodes: List[NLPNode]
          An artificial root is automatically added to the front of the node list.
        """
        self.nodes = [NLPToken.root()]
        if nodes:
            self.nodes.extend(nodes)

    def __next__(self):
        try:
            return next(self._iter)
        except StopIteration:
            raise StopIteration

    def __iter__(self):
        self._iter = islice(self.nodes, 1, len(self.nodes))
        return self

    def __str__(self):
        return '\n'.join(map(str, self))

    def __len__(self):
        return len(self.nodes) - 1


class Relation(Enum):

    PARENT = 'p'
    LEFTMOST_CHILD = 'lmc'
    RIGHTMOST_CHILD = 'rmc'
    LEFT_NEAREST_CHILD = 'lnc'
    RIGHT_NEAREST_CHILD = 'rnc'
    LEFTMOST_SIBLING = 'lms'
    RIGHTMOST_SIBLING = 'rms'
    LEFT_NEAREST_SIBLING = 'lns'
    RIGHT_NEAREST_SIBLING = 'rns'

    GRANDPARENT = 'gp'
    SND_LEFTMOST_CHILD = 'lmc2'
    SND_RIGHTMOST_CHILD = 'rmc2'
    SND_LEFT_NEAREST_CHILD = 'lnc2'
    SND_RIGHT_NEAREST_CHILD = 'rnc2'
    SND_LEFTMOST_SIBLING = 'lms2'
    SND_RIGHTMOST_SIBLING = 'rms2'
    SND_LEFT_NEAREST_SIBLING = 'lns2'
    SND_RIGHT_NEAREST_SIBLING = 'rns2'




# THIS FILE IS GENERATED FROM SCIPY SETUP.PY
short_version = '0.1.10'
version = '0.1.10'
full_version = '0.1.10'
git_revision = '41f20c35c008a20793efa7fa63f3a7233a53c63c'
release = True
if not release:
    version = full_version



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================


from elit.dev.template.lexicon import NLPLexiconMapper
# from fasttext.model import WordVectorModel
from gensim.models.keyedvectors import KeyedVectors

import unittest
import os
#
# __author__ = 'Jinho D. Choi'
#
# class LexiconTest(unittest.TestCase):
#
#     def setUp(self):
#         self.w2v_file = os.path.join(os.path.abspath(os.path.dirname(__file__)),
#                                      './../elit/resources/sample/sample.w2v.bin')
#         self.word_embeddings = KeyedVectors.load_word2vec_format(self.w2v_file, binary=True)
#
#
#     def test_lexicon(self):
#         lexicon = NLPLexiconMapper(self.word_embeddings)
#         print(lexicon.w2v)
#
# if __name__ == '__main__':
#     unittest.main()



#!/usr/bin/env python

from __future__ import division, print_function
import os
import sys
import subprocess

from setuptools import setup, find_packages, Extension
import setuptools
from setuptools.command.build_ext import build_ext

if sys.version_info[:2] < (3, 4):
    raise RuntimeError("Python version >= 3.4 required.")

# How mature is this project? Common values are
#   3 - Alpha
#   4 - Beta
#   5 - Production/Stable
CLASSIFIERS = """\
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: Apache Software License
Programming Language :: C
Programming Language :: Python :: 3 :: Only
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Topic :: Scientific/Engineering :: Artificial Intelligence
Topic :: Scientific/Engineering :: Information Analysis
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""

MAJOR = 0
MINOR = 1
MICRO = 10
ISRELEASED = True
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)

EXCLUDE_FROM_PACKAGES = ['']

# Return the git revision as a string
def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH', 'HOME']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION


def get_version_info():
    # Adding the git rev number needs to be done inside write_version_py(),
    # otherwise the import of numpy.version messes up the build under Python 3.
    FULLVERSION = VERSION
    if os.path.exists('.git'):
        GIT_REVISION = git_version()
    elif os.path.exists('elit/version.py'):
        # must be a source distribution, use existing version file
        try:
            from elit.version import git_revision as GIT_REVISION
        except ImportError:
            raise ImportError("Unable to import git_revision. Try removing " \
                              "elit/version.py and the build directory " \
                              "before building.")
    else:
        GIT_REVISION = "Unknown"

    if not ISRELEASED:
        FULLVERSION += '.dev0+' + GIT_REVISION[:7]

    return FULLVERSION, GIT_REVISION


def write_version_py(filename='elit/version.py'):
    cnt = """
# THIS FILE IS GENERATED FROM SCIPY SETUP.PY
short_version = '%(version)s'
version = '%(version)s'
full_version = '%(full_version)s'
git_revision = '%(git_revision)s'
release = %(isrelease)s
if not release:
    version = full_version
"""
    FULLVERSION, GIT_REVISION = get_version_info()

    a = open(filename, 'w')
    try:
        a.write(cnt % {'version': VERSION,
                       'full_version': FULLVERSION,
                       'git_revision': GIT_REVISION,
                       'isrelease': str(ISRELEASED)})
    finally:
        a.close()


class PybindInclude(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.

    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)


def setup_package():
    write_version_py()

    ext_modules = [
        Extension(
            'elit.tokenizer.english_tokenizer',
            ['cpp/tokenizer/english_tokenizer_py.cpp',
             'cpp/tokenizer/english_tokenizer.cpp',
             'cpp/tokenizer/token_utils.cpp',
             'cpp/string_utils.cpp',
             'cpp/io_utils.cpp'],
            include_dirs=[
                # Path to pybind11 headers
                PybindInclude(),
                PybindInclude(user=True),
                'elit/cpp',
                'elit/cpp/tokenizer'
            ],
            language='c++'
        ),
    ]

    metadata = dict(
        name='elit',
        url='https://github.com/emorynlp/elit',
        download_url='https://github.com/emorynlp/elit/tree/master',
        author='Jinho D. Choi',
        author_email='choi@mathcs.emory.edu',
        description='The Emory Language Information Toolkit (ELIT).',
        license='ALv2',
        package_data={'': ['resources/tokenizer/*.txt',
                           'resources/sample/*']},
        packages=find_packages(exclude=EXCLUDE_FROM_PACKAGES),
        install_requires=[
            'Cython',
            'mxnet',
            'argparse',
            'gensim',
            'fasttext',
            'numpy',
            'keras',
            'tensorflow',
            'h5py',
            'ujson',
            'pybind11>=2.2'
        ],
        tests_require=[
            'nose',
        ],
        test_suite='nose.collector',
        classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
        ext_modules=ext_modules,
        cmdclass={'build_ext': BuildExt}
    )
    metadata['version'] = get_version_info()[0]

    setup(**metadata)


if __name__ == '__main__':
    setup_package()



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================

import numpy as np
from fasttext.model import WordVectorModel
from gensim.models.keyedvectors import KeyedVectors

from elit.dev import structure
from elit.dev.structure import NLPToken

__author__ = 'Jinho D. Choi'


class NLPEmbedding:
    def __init__(self, vsm, key_field, emb_field):
        """
        :param vsm: the vector space model in either the form of Word2Vec or FastText.
        :type vsm: Union[KeyedVectors, WordVectorModel]
        :param key_field: the field in NLPNode (e.g., word, pos) used as the key to retrieve
                the emb_matrix from vsm.
        :type key_field: str
        :param emb_field: where the emb_matrix with respect to the key is saved in NLPNode.
        :type emb_field: str
        """
        self.vsm = vsm
        self.key_field = key_field
        self.emb_field = emb_field

        if isinstance(vsm, KeyedVectors):
            vector_size = vsm.syn0.shape[1]
            # root
            np.random.seed(9)
            self.root = np.random.uniform(-.25, .25, (vector_size,)).astype('float32')
            np.random.seed()
            # zero
            self.zero = np.zeros((vector_size,)).astype('float32')
        elif isinstance(vsm, WordVectorModel):
            self.root = np.array(vsm[structure.ROOT_TAG]).astype('float32')
            self.zero = np.array(vsm['']).astype('float32')

    def get(self, node):
        """
        :param node:
        :type node: NLPToken
        :return: the emb_matrix of the specific node with respect to the key_field.
        :rtype: np.array
        """
        if node is None:
            return self.zero
        if node.token_id == 0:
            return self.root
        if hasattr(node, self.emb_field):
            return getattr(node, self.emb_field)

        f = getattr(node, self.key_field)
        emb = None

        if isinstance(self.vsm, KeyedVectors):
            vocab = self.vsm.vocab.get(f, None)
            emb = self.zero if vocab is None else self.vsm.syn0[vocab.index]
        elif isinstance(self.vsm, WordVectorModel):
            emb = np.array(self.vsm[f]).astype('float32')

        setattr(node, self.emb_field, emb)
        return emb


class NLPLexiconMapper:
    def __init__(self, w2v=None, f2v=None):
        """
        :param w2v: KeyedVectors.load_word2vec_format('*.bin').
        :type KeyedVectors
        :param f2v: f2v.load_model('*.bin').
        :type WordVectorModel
        """
        self.w2v = NLPEmbedding(w2v, 'word', 'w2v') if w2v else None
        self.f2v = NLPEmbedding(f2v, 'word', 'f2v') if f2v else None



# ========================================================================
# Copyright 2017 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
from bisect import *

__author__ = 'Jinho D. Choi'


def bisect_index(a, x, lo=0, hi=None) -> int:
    """
    :param a: a sorted list.
    :param x: the element to search for.
    :param lo: the lower-bound for search (inclusive).
    :param hi: the upper-bound for search (exclusive).
    :return: the index of the node in the sorted list of tuples if exists; otherwise, -1.
    """
    if hi is None: hi = len(a)
    idx = bisect_left(a, x, lo, hi)
    return idx if idx != len(a) and a[idx] == x else -1


def bisect_remove(a, x, lo=0, hi=None) -> int:
    """
    :param a: a sorted list.
    :param x: the element to search for.
    :param lo: the lower-bound for search (inclusive).
    :param hi: the upper-bound for search (exclusive).
    :return: the index of the removed item in the sorted list of tuples if exists; otherwise, -1.
    """
    idx = bisect_index(a, x, lo, hi)
    if idx >= 0: del a[idx]
    return idx

